#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Reboiler model for solvent columns.

This is a simple model for a reboiler in the case where liquid and vapor phases
have separate property packages, such as the case of solvent columns.

Assumptions:
     * Steady-state only
     * Liquid phase property package has a single phase named Liq
     * Vapor phase property package has a single phase named Vap
     * Liquid and vapor phase properties need not have the same component lists
"""

__author__ = "Andrew Lee, Paul Akula"

# Import Pyomo libraries
from pyomo.environ import (
    check_optimal_termination,
    Constraint,
    Param,
    Reference,
    units as pyunits,
    value,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# Import IDAES cores
import idaes.logger as idaeslog
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    EnergyBalanceType,
    MomentumBalanceType,
    MaterialBalanceType,
    MaterialFlowBasis,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.exceptions import ConfigurationError, InitializationError


_log = idaeslog.getIdaesLogger(__name__)


@declare_process_block_class("SolventReboiler")
class SolventReboilerData(UnitModelBlockData):
    """
    Reboiler unit for solvent column models using separate property packages
    for liquid and vpor phases.

    Unit model to reboil the liquid from the bottom of a solvent column.
    """

    CONFIG = ConfigBlock()
    # TODO: Add dynamics in future
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Equilibrium Reactors do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Equilibrium reactors do not have defined volume, thus
this must be False.""",
        ),
    )
    # TODO : Add boilup ratio back later if needed
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for liquid phase",
            doc="""Property parameter object used to define property calculations
for the liquid phase,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing liquid phase properties",
            doc="""A ConfigBlock with arguments to be passed to liquid phase
property block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "vapor_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for vapor phase",
            doc="""Property parameter object used to define property calculations
for the vapor phase,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "vapor_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing vapor phase properties",
            doc="""A ConfigBlock with arguments to be passed to vapor phase
property block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """Build the model.

        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()

        # Check phase lists match assumptions
        if self.config.vapor_property_package.phase_list != ["Vap"]:
            raise ConfigurationError(
                f"{self.name} SolventReboiler model requires that the vapor "
                f"phase property package have a single phase named 'Vap'"
            )
        if self.config.liquid_property_package.phase_list != ["Liq"]:
            raise ConfigurationError(
                f"{self.name} SolventReboiler model requires that the liquid "
                f"phase property package have a single phase named 'Liq'"
            )

        # Check for at least one common component in component lists
        if not any(
            j in self.config.vapor_property_package.component_list
            for j in self.config.liquid_property_package.component_list
        ):
            raise ConfigurationError(
                f"{self.name} SolventReboiler model requires that the liquid "
                f"and vapor phase property packages have at least one "
                f"common component."
            )

        # ---------------------------------------------------------------------
        # Add Control Volume for the Liquid Phase
        self.liquid_phase = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.liquid_property_package,
            property_package_args=self.config.liquid_property_package_args,
        )

        self.liquid_phase.add_state_blocks(has_phase_equilibrium=True)

        # Separate liquid and vapor phases means that phase equilibrium will
        # be handled at the unit model level, thus has_phase_equilibrium is
        # False, but has_mass_transfer is True.
        self.liquid_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True,
            has_phase_equilibrium=False,
        )

        # Need to include enthalpy transfer term for the mass transfer
        self.liquid_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=True,
            has_enthalpy_transfer=True,
        )

        self.liquid_phase.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # ---------------------------------------------------------------------
        # Add single state block for vapor phase
        tmp_dict = dict(**self.config.vapor_property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = False
        self.vapor_phase = self.config.vapor_property_package.build_state_block(
            self.flowsheet().time, doc="Vapor phase properties", **tmp_dict
        )

        # ---------------------------------------------------------------------
        # Check flow basis is compatible
        # TODO : Could add code to convert flow bases, but not now
        t_init = self.flowsheet().time.first()
        if (
            self.vapor_phase[t_init].get_material_flow_basis()
            != self.liquid_phase.properties_out[t_init].get_material_flow_basis()
        ):
            raise ConfigurationError(
                f"{self.name} vapor and liquid property packages must use the "
                f"same material flow basis."
            )

        # ---------------------------------------------------------------------
        # Add Ports for the reboiler
        self.add_inlet_port(name="inlet", block=self.liquid_phase, doc="Liquid feed")
        self.add_outlet_port(
            name="bottoms", block=self.liquid_phase, doc="Bottoms stream"
        )
        self.add_outlet_port(
            name="vapor_reboil",
            block=self.vapor_phase,
            doc="Vapor stream from reboiler",
        )

        # ---------------------------------------------------------------------
        # Add unit level constraints
        # First, need the union and intersection of component lists
        all_comps = (
            self.vapor_phase.component_list
            | self.liquid_phase.properties_out.component_list
        )
        common_comps = (
            self.vapor_phase.component_list
            & self.liquid_phase.properties_out.component_list
        )

        # Get units for unit conversion
        vunits = self.config.vapor_property_package.get_metadata().get_derived_units
        lunits = self.config.liquid_property_package.get_metadata().get_derived_units
        flow_basis = self.vapor_phase[t_init].get_material_flow_basis()
        if flow_basis == MaterialFlowBasis.molar:
            fb = "flow_mole"
        elif flow_basis == MaterialFlowBasis.mass:
            fb = "flow_mass"
        else:
            raise ConfigurationError(
                f"{self.name} SolventReboiler only supports mass or molar "
                f"basis for MaterialFlowBasis."
            )

        if any(j not in common_comps for j in self.vapor_phase.component_list):
            # We have non-condensable components present, need zero-flow param
            self.zero_flow_param = Param(
                mutable=True, default=1e-8, units=vunits("flow_mole")
            )

        # Material balances
        def rule_material_balance(blk, t, j):
            if j in common_comps:
                # Component is in equilibrium
                # Mass transfer equals vapor flowrate
                return -blk.liquid_phase.mass_transfer_term[
                    t, "Liq", j
                ] == pyunits.convert(
                    blk.vapor_phase[t].get_material_flow_terms("Vap", j),
                    to_units=lunits(fb),
                )
            elif j in self.vapor_phase.component_list:
                # Non-condensable component
                # No mass transfer term
                # Set vapor flowrate to an arbitrary small value
                return (
                    blk.vapor_phase[t].get_material_flow_terms("Vap", j)
                    == blk.zero_flow_param
                )
            else:
                # Non-vaporisable component
                # Mass transfer term is zero, no vapor flowrate
                return blk.liquid_phase.mass_transfer_term[t, "Liq", j] == 0 * lunits(
                    fb
                )

        self.unit_material_balance = Constraint(
            self.flowsheet().time,
            all_comps,
            rule=rule_material_balance,
            doc="Unit level material balances",
        )

        # Phase equilibrium constraints
        # For all common components, equate fugacity in vapor and liquid
        def rule_phase_equilibrium(blk, t, j):
            return blk.liquid_phase.properties_out[t].fug_phase_comp[
                "Liq", j
            ] == pyunits.convert(
                blk.vapor_phase[t].fug_phase_comp["Vap", j], to_units=lunits("pressure")
            )

        self.unit_phase_equilibrium = Constraint(
            self.flowsheet().time,
            common_comps,
            rule=rule_phase_equilibrium,
            doc="Unit level phase equilibrium constraints",
        )

        # Temperature equality constraint
        def rule_temperature_balance(blk, t):
            return blk.liquid_phase.properties_out[t].temperature == pyunits.convert(
                blk.vapor_phase[t].temperature, to_units=lunits("temperature")
            )

        self.unit_temperature_equality = Constraint(
            self.flowsheet().time,
            rule=rule_temperature_balance,
            doc="Unit level temperature equality",
        )

        # Unit level energy balance
        # Energy leaving in vapor phase must be equal and opposite to enthalpy
        # transfer from liquid phase
        def rule_energy_balance(blk, t):
            return -blk.liquid_phase.enthalpy_transfer[t] == pyunits.convert(
                blk.vapor_phase[t].get_enthalpy_flow_terms("Vap"),
                to_units=lunits("energy") / lunits("time"),
            )

        self.unit_enthalpy_balance = Constraint(
            self.flowsheet().time,
            rule=rule_energy_balance,
            doc="Unit level enthalpy_balance",
        )

        # Pressure balance constraint
        def rule_pressure_balance(blk, t):
            return blk.liquid_phase.properties_out[t].pressure == pyunits.convert(
                blk.vapor_phase[t].pressure, to_units=lunits("pressure")
            )

        self.unit_pressure_balance = Constraint(
            self.flowsheet().time,
            rule=rule_pressure_balance,
            doc="Unit level pressure balance",
        )

        # Set references to balance terms at unit level
        self.heat_duty = Reference(self.liquid_phase.heat[:])

        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != MomentumBalanceType.none
        ):
            self.deltaP = Reference(self.liquid_phase.deltaP[:])

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        common_comps = (
            self.vapor_phase.component_list
            & self.liquid_phase.properties_out.component_list
        )

        for (t, j), v in self.unit_material_balance.items():
            if j in common_comps:
                iscale.constraint_scaling_transform(
                    v,
                    iscale.get_scaling_factor(
                        self.liquid_phase.mass_transfer_term[t, "Liq", j],
                        default=1,
                        warning=True,
                    ),
                )
            elif j in self.vapor_phase.component_list:
                iscale.constraint_scaling_transform(v, value(1 / self.zero_flow_param))
            else:
                pass  # no need to scale this constraint

        for (t, j), v in self.unit_phase_equilibrium.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.liquid_phase.properties_out[t].fug_phase_comp["Liq", j],
                    default=1,
                    warning=True,
                ),
            )

        for t, v in self.unit_temperature_equality.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.liquid_phase.properties_out[t].temperature,
                    default=1,
                    warning=True,
                ),
            )

        for t, v in self.unit_enthalpy_balance.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.liquid_phase.enthalpy_transfer[t], default=1, warning=True
                ),
            )

        for t, v in self.unit_pressure_balance.items():
            iscale.constraint_scaling_transform(
                v,
                iscale.get_scaling_factor(
                    self.liquid_phase.properties_out[t].pressure,
                    default=1,
                    warning=True,
                ),
            )

    def initialize(
        blk,
        liquid_state_args=None,
        vapor_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for solvent reboiler unit model.

        Keyword Arguments:
            liquid_state_args : a dict of arguments to be passed to the
                liquid property packages to provide an initial state for
                initialization (see documentation of the specific property
                package) (default = none).
            vapor_state_args : a dict of arguments to be passed to the
                vapor property package to provide an initial state for
                initialization (see documentation of the specific property
                package) (default = none).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default IDAES solver)

        Returns:
            None
        """
        if optarg is None:
            optarg = {}

        # Check DOF
        if degrees_of_freedom(blk) != 0:
            raise InitializationError(
                f"{blk.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(blk)}"
            )

        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        solverobj = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize liquid phase control volume block
        flags = blk.liquid_phase.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=liquid_state_args,
            hold_state=True,
        )

        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Initialize vapor phase state block
        if vapor_state_args is None:
            t_init = blk.flowsheet().time.first()
            vapor_state_args = {}
            vap_state_vars = blk.vapor_phase[t_init].define_state_vars()

            liq_state = blk.liquid_phase.properties_out[t_init]

            # Check for unindexed state variables
            for sv in vap_state_vars:
                if "flow" in sv:
                    # Flow variable, assume 10% vaporization
                    if "phase_comp" in sv:
                        # Flow is indexed by phase and component
                        vapor_state_args[sv] = {}
                        for p, j in vap_state_vars[sv]:
                            if j in liq_state.component_list:
                                vapor_state_args[sv][p, j] = 0.1 * value(
                                    getattr(liq_state, sv)[p, j]
                                )
                            else:
                                vapor_state_args[sv][p, j] = 1e-8
                    elif "comp" in sv:
                        # Flow is indexed by component
                        vapor_state_args[sv] = {}
                        for j in vap_state_vars[sv]:
                            if j in liq_state.component_list:
                                vapor_state_args[sv][j] = 0.1 * value(
                                    getattr(liq_state, sv)[j]
                                )
                            else:
                                vapor_state_args[sv][j] = 1e-8
                    elif "phase" in sv:
                        # Flow is indexed by phase
                        vapor_state_args[sv] = {}
                        for p in vap_state_vars[sv]:
                            vapor_state_args[sv][p] = 0.1 * value(
                                getattr(liq_state, sv)["Liq"]
                            )
                    else:
                        vapor_state_args[sv] = 0.1 * value(getattr(liq_state, sv))
                elif "mole_frac" in sv:
                    vapor_state_args[sv] = {}
                    if "phase" in sv:
                        # Variable is indexed by phase and component
                        for p, j in vap_state_vars[sv].keys():
                            if j in liq_state.component_list:
                                vapor_state_args[sv][p, j] = value(
                                    liq_state.fug_phase_comp["Liq", j]
                                    / liq_state.pressure
                                )
                            else:
                                vapor_state_args[sv][p, j] = 1e-8
                    else:
                        for j in vap_state_vars[sv].keys():
                            if j in liq_state.component_list:
                                vapor_state_args[sv][j] = value(
                                    liq_state.fug_phase_comp["Liq", j]
                                    / liq_state.pressure
                                )
                            else:
                                vapor_state_args[sv][j] = 1e-8
                else:
                    vapor_state_args[sv] = value(getattr(liq_state, sv))

        blk.vapor_phase.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=vapor_state_args,
            hold_state=False,
        )

        init_log.info_high("Initialization Step 2 Complete.")
        # ---------------------------------------------------------------------
        # Solve unit model
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solverobj.solve(blk, tee=slc.tee)

        init_log.info_high(
            "Initialization Step 3 {}.".format(idaeslog.condition(results))
        )

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.liquid_phase.release_state(flags, outlvl)

        if not check_optimal_termination(results):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info("Initialization Complete: {}".format(idaeslog.condition(results)))

    # TODO : performance and stream table methods



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Steam turbine inlet stage model.  This model is based on:

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

__Author__ = "John Eslick"

from pyomo.environ import Var, sqrt, value, units as pyunits
from idaes.core import declare_process_block_class
from idaes.models_extra.power_generation.unit_models.helm.turbine import (
    HelmIsentropicTurbineData,
)
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale


_log = idaeslog.getLogger(__name__)


@declare_process_block_class(
    "HelmTurbineInletStage",
    doc="Inlet stage steam turbine model",
)
class HelmTurbineInletStageData(HelmIsentropicTurbineData):
    CONFIG = HelmIsentropicTurbineData.CONFIG()

    def build(self):
        super().build()

        self.flow_coeff = Var(
            self.flowsheet().time,
            initialize=1.053 / 3600.0,
            doc="Turbine flow coefficient [kg*C^0.5/Pa/s]",
            units=pyunits.kg * pyunits.K**0.5 / pyunits.Pa / pyunits.s,
        )
        self.blade_reaction = Var(initialize=0.9, doc="Blade reaction parameter")
        self.blade_velocity = Var(
            initialize=110.0,
            doc="Design blade velocity [m/s]",
            units=pyunits.m / pyunits.s,
        )
        self.eff_nozzle = Var(
            initialize=0.95,
            bounds=(0.0, 1.0),
            doc="Nozzel efficiency (typically 0.90 to 0.95)",
        )
        self.efficiency_mech = Var(initialize=1.0, doc="Turbine mechanical efficiency")

        self.eff_nozzle.fix()
        self.blade_reaction.fix()
        self.flow_coeff.fix()
        self.blade_velocity.fix()
        self.efficiency_mech.fix()
        self.efficiency_isentropic.unfix()
        self.ratioP[:] = 0.9  # make sure these have a number value
        self.deltaP[:] = 0  #   to avoid an error later in initialize

        @self.Expression(
            self.flowsheet().time,
            doc="Entering steam velocity calculation [m/s]",
        )
        def steam_entering_velocity(b, t):
            # 1.414 = 44.72/sqrt(1000) for SI if comparing to Liese (2014),
            # b.delta_enth_isentropic[t] = -(hin - hiesn), the mw converts
            # enthalpy to a mass basis
            return 1.414 * sqrt(
                (b.blade_reaction - 1)
                * b.delta_enth_isentropic[t]
                * self.eff_nozzle
                / b.control_volume.properties_in[t].mw
            )

        @self.Expression(self.flowsheet().time, doc="Efficiency expression")
        def efficiency_isentropic_expr(b, t):
            Vr = b.blade_velocity / b.steam_entering_velocity[t]
            R = b.blade_reaction
            return 2 * Vr * ((sqrt(1 - R) - Vr) + sqrt((sqrt(1 - R) - Vr) ** 2 + R))

        @self.Constraint(self.flowsheet().time, doc="Equation: Turbine inlet flow")
        def inlet_flow_constraint(b, t):
            # Some local vars to make the equation more readable
            g = b.control_volume.properties_in[t].heat_capacity_ratio
            mw = b.control_volume.properties_in[t].mw
            flow = b.control_volume.properties_in[t].flow_mol
            Tin = b.control_volume.properties_in[t].temperature
            cf = b.flow_coeff[t]
            Pin = b.control_volume.properties_in[t].pressure
            Pratio = b.ratioP[t]

            return flow**2 * mw**2 * Tin == (
                cf**2
                * Pin**2
                * g
                / (g - 1)
                * (Pratio ** (2.0 / g) - Pratio ** ((g + 1) / g))
            )

        @self.Constraint(self.flowsheet().time, doc="Equation: Efficiency")
        def efficiency_correlation(b, t):
            return b.efficiency_isentropic[t] == b.efficiency_isentropic_expr[t]

        @self.Expression(self.flowsheet().time, doc="Thermodynamic power [J/s]")
        def power_thermo(b, t):
            return b.control_volume.work[t]

        @self.Expression(self.flowsheet().time, doc="Shaft power [J/s]")
        def power_shaft(b, t):
            return b.power_thermo[t] * b.efficiency_mech

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        calculate_cf=False,
    ):
        """
        Initialize the inlet turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves. This initialization uses a
        flow value guess, so some reasonable flow guess should be specified prior
        to initialization.

        Args:
            outlvl (int): Amount of output (0 to 3) 0 is lowest
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
            calculate_cf (bool): If True, use the flow and pressure ratio to
                calculate the flow coefficient.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # sp is what to save to make sure state after init is same as the start
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # Setup for initialization step 1
        self.inlet_flow_constraint.deactivate()
        self.efficiency_correlation.deactivate()
        self.eff_nozzle.fix()
        self.blade_reaction.fix()
        self.flow_coeff.fix()
        self.blade_velocity.fix()
        self.inlet.fix()
        self.outlet.unfix()

        for t in self.flowsheet().time:
            self.efficiency_isentropic[t] = 0.9
        super().initialize_build(outlvl=outlvl, solver=solver, optarg=optarg)

        # Free eff_isen and activate special constraint
        self.inlet_flow_constraint.activate()
        self.efficiency_correlation.activate()

        if calculate_cf:
            self.ratioP.fix()
            self.flow_coeff.unfix()

            for t in self.flowsheet().time:
                g = self.control_volume.properties_in[t].heat_capacity_ratio
                mw = self.control_volume.properties_in[t].mw
                flow = self.control_volume.properties_in[t].flow_mol
                Tin = self.control_volume.properties_in[t].temperature
                Pin = self.control_volume.properties_in[t].pressure
                Pratio = self.ratioP[t]
                self.flow_coeff[t].value = value(
                    flow
                    * mw
                    * sqrt(
                        Tin
                        / (
                            g
                            / (g - 1)
                            * (Pratio ** (2.0 / g) - Pratio ** ((g + 1) / g))
                        )
                    )
                    / Pin
                )

        # Create solver
        slvr = get_solver(solver, optarg)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = slvr.solve(self, tee=slc.tee)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))
        # reload original spec
        if calculate_cf:
            cf = {}
            for t in self.flowsheet().time:
                cf[t] = value(self.flow_coeff[t])

        from_json(self, sd=istate, wts=sp)
        if calculate_cf:
            # cf was probably fixed, so will have to set the value again here
            # if you ask for it to be calculated.
            for t in self.flowsheet().time:
                self.flow_coeff[t] = cf[t]

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for t, c in self.inlet_flow_constraint.items():
            s = (
                iscale.get_scaling_factor(self.control_volume.properties_in[t].flow_mol)
                ** 2
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for the direct air capture costing model.
"""

__author__ = "Alex Noring"

import pytest

from pyomo.environ import check_optimal_termination, ConcreteModel, value

from idaes.core import FlowsheetBlock
from idaes.models_extra.temperature_swing_adsorption import (
    FixedBedTSA0D,
    FixedBedTSA0DInitializer,
    Adsorbent,
    TransformationScheme,
    SteamCalculationType,
)
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock
from idaes.models_extra.temperature_swing_adsorption.costing.dac_costing import (
    get_dac_costing,
    print_dac_costing,
    dac_costing_summary,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.integration
class TestElectricBoilerCosting:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.compressor_props = FlueGasParameterBlock(components=["N2", "CO2"])
        m.fs.steam_props = iapws95.Iapws95ParameterBlock()

        m.fs.unit = FixedBedTSA0D(
            adsorbent=Adsorbent.zeolite_13x,
            number_of_beds=600,
            transformation_method="dae.collocation",
            transformation_scheme=TransformationScheme.lagrangeRadau,
            finite_elements=20,
            collocation_points=6,
            compressor=True,
            compressor_properties=m.fs.compressor_props,
            steam_calculation=SteamCalculationType.rigorous,
            steam_properties=m.fs.steam_props,
        )

        m.fs.unit.inlet.flow_mol_comp[0, "H2O"].fix(0)
        m.fs.unit.inlet.flow_mol_comp[0, "CO2"].fix(40)
        m.fs.unit.inlet.flow_mol_comp[0, "N2"].fix(99960)
        m.fs.unit.inlet.flow_mol_comp[0, "O2"].fix(0)
        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(100000)

        m.fs.unit.temperature_desorption.fix(470)
        m.fs.unit.temperature_adsorption.fix(310)
        m.fs.unit.temperature_heating.fix(500)
        m.fs.unit.temperature_cooling.fix(300)
        m.fs.unit.bed_diameter.fix(4)
        m.fs.unit.bed_height.fix(8)
        m.fs.unit.compressor.unit.efficiency_isentropic.fix(0.8)

        iscale.calculate_scaling_factors(m)

        initializer = FixedBedTSA0DInitializer()
        initializer.initialize(m.fs.unit)

        get_dac_costing(m.fs.unit, "electric_boiler")

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_build(self, model):
        assert hasattr(model.fs, "costing")
        assert hasattr(model.fs.costing, "total_TPC")
        assert hasattr(model.fs.costing, "total_fixed_OM_cost")
        assert hasattr(model.fs.costing, "total_variable_OM_cost")

        assert number_variables(model) == 3007
        assert number_total_constraints(model) == 2976
        assert number_unused_variables(model) == 12

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, model):
        solver.options.bound_push = 1e-6
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, model):
        # CO2 rich outlet
        assert pytest.approx(79.1955, abs=1e-4) == value(
            model.fs.costing.annualized_cost
        )
        assert pytest.approx(27.1399, abs=1e-4) == value(
            model.fs.costing.total_fixed_OM_cost
        )
        assert pytest.approx(1090.57, abs=1e-2) == value(
            model.fs.costing.total_variable_OM_cost[0]
        )
        assert pytest.approx(0.317544, abs=1e-6) == value(
            model.fs.costing.cost_of_capture
        )

    @pytest.mark.ui
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_report(self, model):
        print_dac_costing(model.fs.unit)
        dac_costing_summary(model.fs.unit)


@pytest.mark.integration
class TestRetrofitNgccCosting:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.compressor_props = FlueGasParameterBlock(components=["N2", "CO2"])
        m.fs.steam_props = iapws95.Iapws95ParameterBlock()

        m.fs.unit = FixedBedTSA0D(
            adsorbent=Adsorbent.zeolite_13x,
            number_of_beds=600,
            transformation_method="dae.collocation",
            transformation_scheme=TransformationScheme.lagrangeRadau,
            finite_elements=20,
            collocation_points=6,
            compressor=True,
            compressor_properties=m.fs.compressor_props,
            steam_calculation=SteamCalculationType.rigorous,
            steam_properties=m.fs.steam_props,
        )

        m.fs.unit.inlet.flow_mol_comp[0, "H2O"].fix(0)
        m.fs.unit.inlet.flow_mol_comp[0, "CO2"].fix(40)
        m.fs.unit.inlet.flow_mol_comp[0, "N2"].fix(99960)
        m.fs.unit.inlet.flow_mol_comp[0, "O2"].fix(0)
        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(100000)

        m.fs.unit.temperature_desorption.fix(470)
        m.fs.unit.temperature_adsorption.fix(310)
        m.fs.unit.temperature_heating.fix(500)
        m.fs.unit.temperature_cooling.fix(300)
        m.fs.unit.bed_diameter.fix(4)
        m.fs.unit.bed_height.fix(8)
        m.fs.unit.compressor.unit.efficiency_isentropic.fix(0.8)

        iscale.calculate_scaling_factors(m)

        initializer = FixedBedTSA0DInitializer()
        initializer.initialize(m.fs.unit)

        get_dac_costing(m.fs.unit, "retrofit_ngcc")

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_build(self, model):
        assert hasattr(model.fs, "costing")
        assert hasattr(model.fs.costing, "total_TPC")
        assert hasattr(model.fs.costing, "total_fixed_OM_cost")
        assert hasattr(model.fs.costing, "total_variable_OM_cost")

        assert number_variables(model) == 2961
        assert number_total_constraints(model) == 2930
        assert number_unused_variables(model) == 12

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, model):
        solver.options.bound_push = 1e-6
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, model):
        # CO2 rich outlet
        assert pytest.approx(81.1308, abs=1e-4) == value(
            model.fs.costing.annualized_cost
        )
        assert pytest.approx(27.8940, abs=1e-4) == value(
            model.fs.costing.total_fixed_OM_cost
        )
        assert pytest.approx(1020.94, abs=1e-2) == value(
            model.fs.costing.total_variable_OM_cost[0]
        )
        assert pytest.approx(0.300183, abs=1e-6) == value(
            model.fs.costing.cost_of_capture
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

# Changing existing config block attributes
# pylint: disable=protected-access

import pyomo.environ as pyo
from pyomo.common.config import In
from idaes.core import declare_process_block_class
from idaes.models_extra.power_generation.unit_models.balance import BalanceBlockData
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
import idaes.models.properties.helmholtz.helmholtz as hltz
from idaes.models.properties.helmholtz.helmholtz import (
    HelmholtzThermoExpressions as ThermoExpr,
)
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale


_log = idaeslog.getLogger(__name__)


def _assert_properties(pb):
    """Assert that the properties parameter block conforms to the requirements"""
    try:
        assert isinstance(pb, hltz.HelmholtzParameterBlockData)
        assert pb.config.phase_presentation in {
            hltz.PhaseType.MIX,
            hltz.PhaseType.L,
            hltz.PhaseType.G,
        }
        assert pb.config.state_vars == hltz.StateVars.PH
    except AssertionError:
        _log.error(
            "helm.HelmIsentropicTurbine requires a Helmholtz EOS with "
            "a single or mixed phase and pressure-enthalpy state vars."
        )
        raise


@declare_process_block_class("HelmIsentropicTurbine")
class HelmIsentropicTurbineData(BalanceBlockData):
    """
    Basic isentropic 0D turbine model.  This inherits the heater block to get
    a lot of unit model boilerplate and the mass balance, enegy balance and
    pressure equations.  This model is intended to be used only with Helmholtz
    EOS property packages in mixed or single phase mode with P-H state vars.

    Since this inherits BalanceBlockData, and only operates in steady-state or
    pseudo-steady-state (for dynamic models) the following mass, energy and
    pressure equations are implicitly written.

    1) Mass Balance:
        0 = flow_mol_in[t] - flow_mol_out[t]
    2) Energy Balance:
        0 = (flow_mol[t]*h_mol[t])_in - (flow_mol[t]*h_mol[t])_out + Q_in + W_in
    3) Pressure:
        0 = P_in[t] + deltaP[t] - P_out[t]
    """

    CONFIG = BalanceBlockData.CONFIG()
    # For dynamics assume pseudo-steady-state
    CONFIG.dynamic = False
    CONFIG.get("dynamic")._default = False
    CONFIG.get("dynamic")._domain = In([False])
    CONFIG.has_holdup = False
    CONFIG.get("has_holdup")._default = False
    CONFIG.get("has_holdup")._domain = In([False])
    # Rest of config to make this function like a turbine
    CONFIG.has_pressure_change = True
    CONFIG.get("has_pressure_change")._default = True
    CONFIG.get("has_pressure_change")._domain = In([True])
    CONFIG.has_work_transfer = True
    CONFIG.get("has_work_transfer")._default = True
    CONFIG.get("has_work_transfer")._domain = In([True])
    CONFIG.has_heat_transfer = False
    CONFIG.get("has_heat_transfer")._default = False
    CONFIG.get("has_heat_transfer")._domain = In([False])

    def build(self):
        """
        Add model equations to the unit model.  This is called by a default block
        construnction rule when the unit model is created.
        """
        super().build()  # Basic unit model build/read config
        config = self.config  # shorter config pointer

        # The thermodynamic expression writer object, te, writes expressions
        # including external function calls to calculate thermodynamic quantities
        # from a set of state variables.
        _assert_properties(config.property_package)
        te = ThermoExpr(blk=self, parameters=config.property_package)

        eff = self.efficiency_isentropic = pyo.Var(
            self.flowsheet().time, initialize=0.9, doc="Isentropic efficiency"
        )
        eff.fix()

        pratio = self.ratioP = pyo.Var(
            self.flowsheet().time,
            initialize=0.7,
            doc="Ratio of outlet to inlet pressure",
        )

        # Some shorter references to property blocks
        properties_in = self.control_volume.properties_in
        properties_out = self.control_volume.properties_out

        @self.Expression(self.flowsheet().time, doc="Outlet isentropic enthalpy")
        def h_is(b, t):
            return te.h(s=properties_in[t].entr_mol, p=properties_out[t].pressure)

        @self.Expression(self.flowsheet().time, doc="Isentropic enthalpy change")
        def delta_enth_isentropic(b, t):
            return self.h_is[t] - properties_in[t].enth_mol

        @self.Expression(self.flowsheet().time, doc="Isentropic work")
        def work_isentropic(b, t):
            return properties_in[t].flow_mol * (
                properties_in[t].enth_mol - self.h_is[t]
            )

        @self.Expression(self.flowsheet().time, doc="Outlet enthalpy")
        def h_o(b, t):  # Early access to the outlet enthalpy and work
            return properties_in[t].enth_mol - eff[t] * (
                properties_in[t].enth_mol - self.h_is[t]
            )

        @self.Constraint(self.flowsheet().time)
        def eq_work(b, t):  # Work from energy balance
            return properties_out[t].enth_mol == self.h_o[t]

        @self.Constraint(self.flowsheet().time)
        def eq_pressure_ratio(b, t):
            return pratio[t] * properties_in[t].pressure == properties_out[t].pressure

        @self.Expression(self.flowsheet().time)
        def work_mechanical(b, t):
            return b.control_volume.work[t]

    def _get_performance_contents(self, time_point=0):
        """This returns a dictionary of quntities to be used in IDAES unit model
        report generation routines.
        """
        pc = super()._get_performance_contents(time_point=time_point)
        return pc

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        For simplicity this initialization requires you to set values for the
        efficiency, inlet, and one of pressure ratio, pressure change or outlet
        pressure.
        """
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        slvr = get_solver(solver, optarg)

        # Store original specification so initialization doesn't change the model
        # This will only restore the values of variables that were originally fixed
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)
        # Check for alternate pressure specs
        for t in self.flowsheet().time:
            if self.outlet.pressure[t].fixed:
                self.ratioP[t] = pyo.value(
                    self.outlet.pressure[t] / self.inlet.pressure[t]
                )
            elif self.control_volume.deltaP[t].fixed:
                self.ratioP[t] = pyo.value(
                    (self.control_volume.deltaP[t] + self.inlet.pressure[t])
                    / self.inlet.pressure[t]
                )
        # Fix the variables we base the initialization on and free the rest.
        # This requires good values to be provided for pressure, efficiency,
        # and inlet conditions, but it is simple and reliable.
        self.inlet.fix()
        self.outlet.unfix()
        self.ratioP.fix()
        self.deltaP.unfix()
        self.efficiency_isentropic.fix()
        for t in self.flowsheet().time:
            self.outlet.pressure[t] = pyo.value(self.inlet.pressure[t] * self.ratioP[t])
            self.deltaP[t] = pyo.value(self.outlet.pressure[t] - self.inlet.pressure[t])

            self.outlet.enth_mol[t] = pyo.value(self.h_o[t])
            self.control_volume.work[t] = pyo.value(
                self.inlet.flow_mol[t] * self.inlet.enth_mol[t]
                - self.outlet.flow_mol[t] * self.outlet.enth_mol[t]
            )
            self.outlet.flow_mol[t] = pyo.value(self.inlet.flow_mol[t])
        # Solve the model (should be already solved from above)
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            slvr.solve(self, tee=slc.tee)
        from_json(self, sd=istate, wts=sp)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for t, c in self.eq_pressure_ratio.items():
            s = iscale.get_scaling_factor(self.control_volume.properties_in[t].pressure)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.eq_work.items():
            s = iscale.get_scaling_factor(self.control_volume.work[t])
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Two-dimensional model of a porous, conductive medium, such as electrodes used in
a solid oxide fuel/electrolytic cell. Equations regarding the reactions that
occur at the triple phase boundary are located in the TriplePhaseBoundary class.
Material and heat transfer PDEs are  discretized in both x and z directions using
a finite volume method. Quantities are interpolated using a centered-difference
scheme. Presently only conventional diffusion is considered, because the Knudsen
number is much less than one for reactive species (H2, H2O, and O2). The energy
balance considers thermal condition from the solid phase, enthalpy flows from
gas species, and resistive heating from the current flowing through the medium.

If variables are not provided for boundary conditions at the ``x=0`` or ``x=1``
edges, new variables will be created for those terms. No-flux boundary conditions
are used along the ``z=0``and ``z=1`` edges.

Boundary variables:
    - ``temperature_deviation_x0[t, iz]``
    - ``temperature_deviation_x1[t, iz]``
    - ``heat_flux_x0[t, iz]``
    - ``heat_flux_x1[t, iz]``
    - ``conc_mol_comp_deviation_x0[t, iz]``
    - ``conc_mol_comp_deviation_x1[t, iz]``
    - ``material_flux_x0[t, iz]``
    - ``material_flux_x1[t, iz]``

Instances of ``Var`` that must be fixed:
    - ``length_x``: Thickness of porous conductive slab
    - ``porosity``: Porosity or void fraction of slab
    - ``tortuosity``: Tortuosity of slab
    - ``solid_heat_capacity``: Specific heat capacity of solid phase of slab on a mass-basis (i.e., in J/(kg * K))
    - ``solid_density``: Mass density of the solid component of the slab. Note that literature
      values that include void areas must be corrected by dividing by one minus porosity
    - ``solid_thermal_conductivity``: Thermal conductivity of solid phase of slab. Again, literature
      values may need to be divided by one minus porosity
    - ``resistivity_log_preexponential_factor``: Natural logarithm of resistivity preexponential factor in ohm * m.
      Again, literature values may need to be divided by one minus porosity
    - ``resistivity_thermal_exponent_dividend``: Parameter divided by temperature in resistivity equation, in K.
      Would be something like (reduced) activation energy, but it can be both negative and positive.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "John Eslick, Douglas Allan"

from pyomo.common.config import ConfigValue, Bool, ListOf
from pyomo.dae import DerivativeVar
import pyomo.environ as pyo


from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.constants import Constants
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
from idaes.models_extra.power_generation.unit_models.soc_submodels.common import (
    _set_if_unfixed,
    _gas_species_list,
    _element_list,
    _element_dict,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

import idaes.logger as idaeslog


@declare_process_block_class("PorousConductiveSlab")
class PorousConductiveSlabData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "has_gas_holdup",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, make holdup terms corresponding to gas phase",
        ),
    )
    CONFIG.declare(
        "control_volume_xfaces",
        ConfigValue(
            domain=ListOf(float),
            description="List containing coordinates of control volume faces "
            "in x direction. Coordinates must start with zero, be strictly "
            "increasing, and end with one",
        ),
    )
    CONFIG.declare(
        "component_list",
        ConfigValue(
            domain=common._SubsetOf(_gas_species_list), description="List of components"
        ),
    )
    CONFIG.declare(
        "conc_mol_comp_ref",
        ConfigValue(
            default=None,
            description="Variable for the component concentration in bulk channel ",
        ),
    )
    CONFIG.declare(
        "dconc_mol_comp_refdt",
        ConfigValue(
            default=None,
            description="Variable for time derivative of the "
            "component concentration in the bulk channel",
        ),
    )
    CONFIG.declare(
        "voltage_drop_custom",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, add voltage_drop_custom Var to be connected to degradation models",
        ),
    )
    common._submodel_boilerplate_config(CONFIG)
    common._thermal_boundary_conditions_config(CONFIG, thin=False)
    common._material_boundary_conditions_config(CONFIG, thin=False)

    def build(self):
        super().build()

        # Set up some sets for the space and time indexing
        dynamic = self.config.dynamic
        tset = self.flowsheet().config.time
        self.component_list = pyo.Set(
            initialize=self.config.component_list,
            ordered=True,
            doc="Set of all gas-phase components present in submodel",
        )
        # Set up node and face sets and get integer indices for them
        izfaces, iznodes = common._face_initializer(
            self, self.config.control_volume_zfaces, "z"
        )
        ixfaces, ixnodes = common._face_initializer(
            self, self.config.control_volume_xfaces, "x"
        )

        # Space saving aliases
        comps = self.component_list
        include_temp_x_thermo = self.config.include_temperature_x_thermo

        # Electrode thickness AKA length in the x direction is specific to the
        # electrode so local variable here is the only option
        self.length_x = pyo.Var(
            doc="Thickness of the slab (x-direction)",
            units=pyo.units.m,
        )

        if self.config.dynamic:
            # If we're dynamic, the user needs to either provide both the
            # reference concentration and its derivative, or provide neither.
            assert (
                self.config.conc_mol_comp_ref is None
                and self.config.dconc_mol_comp_refdt is None
            ) or (
                self.config.conc_mol_comp_ref is not None
                and self.config.dconc_mol_comp_refdt is not None
            )

        if self.config.has_gas_holdup and not self.config.has_holdup:
            raise ConfigurationError(
                "Creating gas holdup terms while not creating other holdup terms is not supported."
            )

        if self.config.conc_mol_comp_ref is None:
            self.conc_mol_comp_ref = pyo.Var(
                tset,
                iznodes,
                comps,
                doc="Concentration of components in the channel bulk",
                initialize=0.0,
                units=pyo.units.mol / pyo.units.m**3,
            )
            if self.config.dynamic and self.config.has_gas_holdup:
                self.dconc_mol_comp_refdt = DerivativeVar(
                    self.conc_mol_comp_ref,
                    wrt=tset,
                    doc="Derivative of concentration of components in the channel bulk",
                    initialize=0.0,
                    units=pyo.units.mol / (pyo.units.s * pyo.units.m**3),
                )
            else:
                self.dconc_mol_comp_refdt = pyo.Param(
                    tset,
                    iznodes,
                    comps,
                    doc="Derivative of concentration of components in the channel bulk",
                    initialize=0.0,
                    units=pyo.units.mol / (pyo.units.s * pyo.units.m**3),
                )
        else:
            self.conc_mol_comp_ref = pyo.Reference(self.config.conc_mol_comp_ref)
            if self.config.dconc_mol_comp_refdt.ctype == DerivativeVar:
                self.dconc_mol_comp_refdt = pyo.Reference(
                    self.config.dconc_mol_comp_refdt, ctype=pyo.Var
                )
            else:
                self.dconc_mol_comp_refdt = pyo.Reference(
                    self.config.dconc_mol_comp_refdt
                )
        common._submodel_boilerplate_create_if_none(self)
        common._create_thermal_boundary_conditions_if_none(self, thin=False)
        common._create_material_boundary_conditions_if_none(self, thin=False)

        self.porosity = pyo.Var(
            initialize=0.50, doc="Electrode porosity", units=pyo.units.dimensionless
        )
        self.tortuosity = pyo.Var(
            initialize=2.0, doc="Electrode tortuosity", units=pyo.units.dimensionless
        )

        self.temperature_deviation_x = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            doc="Deviation of temperature at node centers from temperature_z",
            units=pyo.units.K,
            bounds=(-1000, 1000),
        )
        self.conc_mol_comp_deviation_x = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            comps,
            doc="Deviation of component concentration at node centers "
            "from conc_mol_comp_ref",
            units=pyo.units.mol / pyo.units.m**3,
            bounds=(-100, 100),
        )
        self.enth_mol = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            doc="Molar enthalpy at node centers",
            units=pyo.units.J / pyo.units.mol,
        )
        if self.config.has_holdup:
            if self.config.has_gas_holdup:
                self.int_energy_mol = pyo.Var(
                    tset,
                    ixnodes,
                    iznodes,
                    doc="Fluid molar internal energy at node centers",
                    units=pyo.units.J / pyo.units.mol,
                )
                self.int_energy_density = pyo.Var(
                    tset,
                    ixnodes,
                    iznodes,
                    doc="Fluid molar internal energy density at node centers",
                    units=pyo.units.J / pyo.units.m**3,
                )
            self.int_energy_density_solid = pyo.Var(
                tset,
                ixnodes,
                iznodes,
                doc="Internal energy density of solid electrode",
                units=pyo.units.J / pyo.units.m**3,
            )
        self.pressure = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            doc="Pressure at node centers",
            units=pyo.units.Pa,
            bounds=(0, None),
        )

        # Assume the the electrode gas phase and solid are same temp
        self.mole_frac_comp = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            comps,
            doc="Component mole fraction at node centers",
            initialize=1 / len(comps),
            bounds=(0, 1),
        )
        self.diff_eff_coeff = pyo.Var(
            tset,
            ixnodes,
            iznodes,
            comps,
            doc="Effective Fick's law diffusion coefficient at node centers",
            initialize=2e-5,
            bounds=(0, None),
            units=pyo.units.m**2 / pyo.units.s,
        )
        self.resistivity_log_preexponential_factor = pyo.Var(
            doc="Logarithm of resistivity preexponential factor in units of ohm*m",
            units=pyo.units.dimensionless,
        )
        self.resistivity_thermal_exponent_dividend = pyo.Var(
            doc="Parameter divided by temperature in exponential", units=pyo.units.K
        )
        if self.config.voltage_drop_custom:
            self.voltage_drop_custom = pyo.Var(
                tset,
                ixnodes,
                iznodes,
                units=pyo.units.volts,
                doc="Custom voltage drop term for degradation modeling",
            )

        # Parameters
        self.solid_heat_capacity = pyo.Var(
            initialize=200,
            doc="Heat capacity of solid part of porous slab (mass basis)",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        self.solid_density = pyo.Var(
            initialize=1000,
            doc="Density of solid part of porous slab",
            units=pyo.units.kg / pyo.units.m**3,
        )
        self.solid_thermal_conductivity = pyo.Var(
            initialize=80,
            doc="Thermal conductivity of solid part of porous slab",
            units=pyo.units.W / pyo.units.m / pyo.units.K,
        )

        # Add time derivative variable if steady state use const 0.
        if dynamic and self.config.has_gas_holdup:
            self.dconc_mol_comp_deviation_xdt = DerivativeVar(
                self.conc_mol_comp_deviation_x,
                wrt=tset,
                initialize=0,
                doc="Component concentration time derivative in deviation " "variable",
            )
        else:
            self.dconc_mol_comp_deviation_xdt = pyo.Param(
                tset,
                ixnodes,
                iznodes,
                comps,
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3 / pyo.units.s,
            )
        # Add time derivative variable if steady state use const 0.
        if dynamic and self.config.has_gas_holdup:
            self.dcedt = DerivativeVar(
                self.int_energy_density,
                wrt=tset,
                initialize=0,
                doc="Internal energy density time derivative",
            )
        else:
            self.dcedt = pyo.Param(
                tset,
                ixnodes,
                iznodes,
                initialize=0,
                units=pyo.units.W / pyo.units.m**3,
            )
        # Add time derivative variable if steady state use const 0.
        if dynamic:
            self.dcedt_solid = DerivativeVar(
                self.int_energy_density_solid,
                wrt=tset,
                initialize=0,
                doc="Internal energy density time derivative",
            )
        else:
            self.dcedt_solid = pyo.Param(
                tset,
                ixnodes,
                iznodes,
                initialize=0,
                units=pyo.units.W / pyo.units.m**3,
            )

        @self.Expression(iznodes)
        def dz(b, iz):
            return b.zfaces.at(iz + 1) - b.zfaces.at(iz)

        @self.Expression(ixnodes)
        def dx(b, ix):
            return b.xfaces.at(ix + 1) - b.xfaces.at(ix)

        @self.Expression(ixnodes, iznodes)
        def node_volume(b, ix, iz):
            return (
                b.length_x[None]
                * b.length_y[None]
                * b.length_z[None]
                * b.dz[iz]
                * b.dx[ix]
            )

        @self.Expression(ixnodes)
        def zface_area(b, ix):
            return b.length_y[None] * b.length_x[None] * b.dx[ix]

        @self.Expression(iznodes)
        def xface_area(b, iz):
            return b.length_y[None] * b.length_z[None] * b.dz[iz]

        @self.Expression(tset, ixnodes, iznodes)
        def temperature(b, t, ix, iz):
            if include_temp_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x[t, ix, iz]
            else:
                return b.temperature_z[t, iz]

        @self.Expression(tset, iznodes, comps)
        def conc_mol_comp_x0(b, t, iz, j):
            return (
                b.conc_mol_comp_ref[t, iz, j] + b.conc_mol_comp_deviation_x0[t, iz, j]
            )

        @self.Expression(tset, ixnodes, iznodes, comps)
        def conc_mol_comp(b, t, ix, iz, j):
            return (
                b.conc_mol_comp_ref[t, iz, j]
                + b.conc_mol_comp_deviation_x[t, ix, iz, j]
            )

        @self.Expression(tset, iznodes, comps)
        def conc_mol_comp_x1(b, t, iz, j):
            return (
                b.conc_mol_comp_ref[t, iz, j] + b.conc_mol_comp_deviation_x1[t, iz, j]
            )

        @self.Expression(tset, ixnodes, iznodes, comps)
        def dconc_mol_compdt(b, t, ix, iz, j):
            return (
                b.dconc_mol_comp_refdt[t, iz, j]
                + b.dconc_mol_comp_deviation_xdt[t, ix, iz, j]
            )

        @self.Expression(tset, ixnodes, iznodes)
        def vol_mol(b, t, ix, iz):
            return (
                Constants.gas_constant
                * b.temperature[t, ix, iz]
                / b.pressure[t, ix, iz]
            )

        @self.Constraint(tset, ixnodes, iznodes, comps)
        def conc_mol_comp_eqn(b, t, ix, iz, i):
            return (
                b.conc_mol_comp[t, ix, iz, i]
                * b.temperature[t, ix, iz]
                * Constants.gas_constant
                == b.pressure[t, ix, iz] * b.mole_frac_comp[t, ix, iz, i]
            )

        @self.Constraint(tset, ixnodes, iznodes)
        def enth_mol_eqn(b, t, ix, iz):
            return b.enth_mol[t, ix, iz] == sum(
                common._comp_enthalpy_expr(b.temperature[t, ix, iz], i)
                * b.mole_frac_comp[t, ix, iz, i]
                for i in comps
            )

        if self.config.has_holdup:
            if self.config.has_gas_holdup:
                # For the vapor phase
                @self.Constraint(tset, ixnodes, iznodes)
                def int_energy_mol_eqn(b, t, ix, iz):
                    return b.int_energy_mol[t, ix, iz] == sum(
                        common._comp_int_energy_expr(b.temperature[t, ix, iz], i)
                        * b.mole_frac_comp[t, ix, iz, i]
                        for i in comps
                    )

                @self.Constraint(tset, ixnodes, iznodes)
                def int_energy_density_eqn(b, t, ix, iz):
                    return (
                        b.int_energy_density[t, ix, iz]
                        == b.int_energy_mol[t, ix, iz] / b.vol_mol[t, ix, iz]
                    )

            @self.Constraint(tset, ixnodes, iznodes)
            def int_energy_density_solid_eqn(b, t, ix, iz):
                return b.int_energy_density_solid[
                    t, ix, iz
                ] == b.solid_heat_capacity * b.solid_density * (
                    b.temperature[t, ix, iz] - 1000 * pyo.units.K
                )

        @self.Constraint(tset, ixnodes, iznodes)
        def mole_frac_comp_eqn(b, t, ix, iz):
            return 1 == sum(b.mole_frac_comp[t, ix, iz, i] for i in comps)

        @self.Constraint(tset, ixnodes, iznodes, comps)
        def diff_eff_coeff_eqn(b, t, ix, iz, i):
            T = b.temperature[t, ix, iz]
            P = b.pressure[t, ix, iz]
            x = b.mole_frac_comp
            bfun = common._binary_diffusion_coefficient_expr
            return b.diff_eff_coeff[t, ix, iz, i] == b.porosity / b.tortuosity * (
                1.0 - x[t, ix, iz, i]
            ) / sum(x[t, ix, iz, j] / bfun(T, P, i, j) for j in comps if i != j)

        @self.Expression(tset, ixfaces, iznodes, comps)
        def dcdx(b, t, ix, iz, i):
            return common._interpolate_2D(
                ic=ix,
                ifaces=ixfaces,
                nodes=b.xnodes,
                faces=b.xfaces,
                phi_func=lambda ixf: b.conc_mol_comp_deviation_x[t, ixf, iz, i]
                / b.length_x,
                phi_bound_0=(
                    b.conc_mol_comp_deviation_x[t, ixnodes.first(), iz, i]
                    - b.conc_mol_comp_deviation_x0[t, iz, i]
                )
                / (b.xnodes.first() - b.xfaces.first())
                / b.length_x,
                phi_bound_1=(
                    b.conc_mol_comp_deviation_x1[t, iz, i]
                    - b.conc_mol_comp_deviation_x[t, ixnodes.last(), iz, i]
                )
                / (b.xfaces.last() - b.xnodes.last())
                / b.length_x,
                derivative=True,
            )

        conc_grad_units = pyo.units.mol / pyo.units.m**4

        @self.Expression(tset, ixnodes, izfaces, comps)
        def dcdz(b, t, ix, iz, i):
            return common._interpolate_2D(
                ic=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda izf: b.conc_mol_comp[t, ix, izf, i] / b.length_z[None],
                phi_bound_0=0 * conc_grad_units,  # solid wall no flux
                phi_bound_1=0 * conc_grad_units,  # solid wall no flux
                derivative=True,
            )

        @self.Expression(tset, ixfaces, iznodes)
        def dTdx(b, t, ix, iz):
            return common._interpolate_2D(
                ic=ix,
                ifaces=ixfaces,
                nodes=b.xnodes,
                faces=b.xfaces,
                phi_func=lambda ixf: b.temperature_deviation_x[t, ixf, iz] / b.length_x,
                phi_bound_0=(
                    b.temperature_deviation_x[t, ixnodes.first(), iz]
                    - b.temperature_deviation_x0[t, iz]
                )
                / (b.xnodes.first() - b.xfaces.first())
                / b.length_x,
                phi_bound_1=(
                    b.temperature_deviation_x1[t, iz]
                    - b.temperature_deviation_x[t, ixnodes.last(), iz]
                )
                / (b.xfaces.last() - b.xnodes.last())
                / b.length_x,
                derivative=True,
            )

        temp_grad_units = pyo.units.K / pyo.units.m

        @self.Expression(tset, ixnodes, izfaces)
        def dTdz(b, t, ix, iz):
            return common._interpolate_2D(
                ic=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda izf: b.temperature[t, ix, izf] / b.length_z[None],
                phi_bound_0=0 * temp_grad_units,
                phi_bound_1=0 * temp_grad_units,
                derivative=True,
            )

        @self.Expression(tset, ixfaces, iznodes, comps)
        def diff_eff_coeff_xfaces(b, t, ix, iz, i):
            return common._interpolate_2D(
                ic=ix,
                ifaces=ixfaces,
                nodes=b.xnodes,
                faces=b.xfaces,
                phi_func=lambda ixf: b.diff_eff_coeff[t, ixf, iz, i],
                # TODO we can probably use conc_x0 and conc_x1 now
                phi_bound_0=b.diff_eff_coeff[
                    t, ixnodes.first(), iz, i
                ],  # use node value
                phi_bound_1=b.diff_eff_coeff[
                    t, ixnodes.last(), iz, i
                ],  # use node value
                derivative=False,
            )

        diff_coeff_units = pyo.units.m**2 / pyo.units.s

        @self.Expression(tset, ixnodes, izfaces, comps)
        def diff_eff_coeff_zfaces(b, t, ix, iz, i):
            return common._interpolate_2D(
                ic=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda izf: b.diff_eff_coeff[t, ix, izf, i],
                phi_bound_0=0 * diff_coeff_units,  # solid wall no flux
                phi_bound_1=0 * diff_coeff_units,  # solid wall no flux
                derivative=False,
            )

        @self.Expression(tset, ixfaces, iznodes)
        def temperature_xfaces(b, t, ix, iz):
            return common._interpolate_2D(
                ic=ix,
                ifaces=ixfaces,
                nodes=b.xnodes,
                faces=b.xfaces,
                phi_func=lambda ixf: b.temperature[t, ixf, iz],
                phi_bound_0=b.temperature_x0[t, iz],
                phi_bound_1=b.temperature_x1[t, iz],
                derivative=False,
            )

        @self.Expression(tset, ixnodes, izfaces)
        def temperature_zfaces(b, t, ix, iz):
            return common._interpolate_2D(
                ic=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda izf: b.temperature[t, ix, izf],
                phi_bound_0=b.temperature[t, ix, iznodes.first()],
                phi_bound_1=b.temperature[t, ix, iznodes.last()],
                derivative=False,
            )

        @self.Expression(tset, ixfaces, iznodes, comps)
        def material_flux_x(b, t, ix, iz, i):
            return -b.dcdx[t, ix, iz, i] * b.diff_eff_coeff_xfaces[t, ix, iz, i]

        @self.Constraint(tset, iznodes, comps)
        def material_flux_x0_eqn(b, t, iz, i):
            return (
                b.material_flux_x[t, ixfaces.first(), iz, i]
                == b.material_flux_x0[t, iz, i]
            )

        @self.Constraint(tset, iznodes, comps)
        def material_flux_x1_eqn(b, t, iz, i):
            return (
                b.material_flux_x[t, ixfaces.last(), iz, i]
                == b.material_flux_x1[t, iz, i]
            )

        @self.Expression(tset, ixnodes, izfaces, comps)
        def material_flux_z(b, t, ix, iz, i):
            return -b.dcdz[t, ix, iz, i] * b.diff_eff_coeff_zfaces[t, ix, iz, i]

        @self.Expression(tset, ixfaces, iznodes)
        def heat_flux_x(b, t, ix, iz):
            return -(1 - b.porosity) * b.solid_thermal_conductivity * b.dTdx[t, ix, iz]

        @self.Expression(tset, ixnodes, izfaces)
        def heat_flux_z(b, t, ix, iz):
            return -(1 - b.porosity) * b.solid_thermal_conductivity * b.dTdz[t, ix, iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x0_eqn(b, t, iz):
            return b.heat_flux_x0[t, iz] == b.heat_flux_x[t, ixfaces.first(), iz]

        @self.Constraint(tset, iznodes)
        def heat_flux_x1_eqn(b, t, iz):
            return b.heat_flux_x1[t, iz] == b.heat_flux_x[t, ixfaces.last(), iz]

        @self.Expression(tset, ixnodes, iznodes)
        def resistivity(b, t, ix, iz):
            return (
                pyo.units.ohm
                * pyo.units.m
                * pyo.exp(
                    b.resistivity_log_preexponential_factor
                    + b.resistivity_thermal_exponent_dividend / b.temperature[t, ix, iz]
                )
            )

        @self.Expression(tset, ixnodes, iznodes)
        def resistance(b, t, ix, iz):
            return (
                b.resistivity[t, ix, iz]
                * b.length_x
                * b.dx[ix]
                / b.xface_area[iz]
                / (1 - b.porosity)
            )

        @self.Expression(tset, iznodes)
        def current(b, t, iz):
            return b.current_density[t, iz] * b.xface_area[iz]

        @self.Expression(tset, ixnodes, iznodes)
        def voltage_drop(b, t, ix, iz):
            if self.config.voltage_drop_custom:
                return (
                    b.current[t, iz] * b.resistance[t, ix, iz]
                    + b.voltage_drop_custom[t, ix, iz]
                )
            else:
                return b.current[t, iz] * b.resistance[t, ix, iz]

        @self.Expression(tset, iznodes)
        def resistance_total(b, t, iz):
            return sum(b.resistance[t, ix, iz] for ix in ixnodes)

        @self.Expression(tset, iznodes)
        def voltage_drop_total(b, t, iz):
            return sum(b.voltage_drop[t, ix, iz] for ix in ixnodes)

        @self.Constraint(tset, ixnodes, iznodes, comps)
        def material_balance_eqn(b, t, ix, iz, i):
            return b.porosity * b.node_volume[ix, iz] * b.dconc_mol_compdt[
                t, ix, iz, i
            ] == b.xface_area[iz] * (
                b.material_flux_x[t, ix, iz, i] - b.material_flux_x[t, ix + 1, iz, i]
            ) + b.zface_area[
                ix
            ] * (
                b.material_flux_z[t, ix, iz, i] - b.material_flux_z[t, ix, iz + 1, i]
            )

        if dynamic and self.config.has_gas_holdup:
            self.material_balance_eqn[tset.first(), :, :, :].deactivate()

        @self.Expression(tset, ixnodes, iznodes)
        def joule_heating(b, t, ix, iz):
            return b.current[t, iz] * b.voltage_drop[t, ix, iz]

        @self.Constraint(tset, ixnodes, iznodes)
        def energy_balance_solid_eqn(b, t, ix, iz):
            return (
                b.node_volume[ix, iz]
                * (
                    b.porosity * b.dcedt[t, ix, iz]
                    + (1 - b.porosity) * b.dcedt_solid[t, ix, iz]
                )
                == b.xface_area[iz]
                * (b.heat_flux_x[t, ix, iz] - b.heat_flux_x[t, ix + 1, iz])
                + b.zface_area[ix]
                * (b.heat_flux_z[t, ix, iz] - b.heat_flux_z[t, ix, iz + 1])
                + b.joule_heating[t, ix, iz]
                # For mass flux heat transfer include exchange with channel
                # probably make little difference, but want to ensure the energy
                # balance closes
                + b.xface_area[iz]
                * sum(
                    b.material_flux_x[t, ix, iz, i]
                    * common._comp_enthalpy_expr(b.temperature_xfaces[t, ix, iz], i)
                    for i in comps
                )
                - b.xface_area[iz]
                * sum(
                    b.material_flux_x[t, ix + 1, iz, i]
                    * common._comp_enthalpy_expr(b.temperature_xfaces[t, ix + 1, iz], i)
                    for i in comps
                )
                + b.zface_area[ix]
                * sum(
                    b.material_flux_z[t, ix, iz, i]
                    * common._comp_enthalpy_expr(b.temperature_zfaces[t, ix, iz], i)
                    for i in comps
                )
                - b.zface_area[ix]
                * sum(
                    b.material_flux_z[t, ix, iz + 1, i]
                    * common._comp_enthalpy_expr(b.temperature_zfaces[t, ix, iz + 1], i)
                    for i in comps
                )
            )

        if dynamic:
            self.energy_balance_solid_eqn[tset.first(), :, :].deactivate()

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        temperature_guess=None,
        pressure_guess=None,
        mole_frac_guess=None,
    ):
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        comps = self.component_list
        # Use conc_x0 and the temperature guess to start filling in initial
        # guess values.
        for t in self.flowsheet().time:
            for iz in self.iznodes:
                for i in comps:
                    _set_if_unfixed(
                        self.conc_mol_comp_deviation_x[t, self.ixnodes.first(), iz, i],
                        self.conc_mol_comp_deviation_x0[t, iz, i],
                    )
                for ix in self.ixnodes:
                    if temperature_guess is not None:
                        _set_if_unfixed(self.temperature_z[t, iz], temperature_guess)
                        _set_if_unfixed(self.temperature_deviation_x[t, ix, iz], 0)
                        _set_if_unfixed(self.temperature_deviation_x0[t, iz], 0)
                        _set_if_unfixed(self.temperature_deviation_x1[t, iz], 0)
                    if pressure_guess is not None:
                        _set_if_unfixed(self.pressure[t, ix, iz], pressure_guess)
                    for i in comps:
                        _set_if_unfixed(self.conc_mol_comp_deviation_x[t, ix, iz, i], 0)
                    mol_dens = pyo.value(
                        sum(self.conc_mol_comp[t, ix, iz, i] for i in comps)
                    )
                    _set_if_unfixed(
                        self.pressure[t, ix, iz],
                        Constants.gas_constant * self.temperature[t, ix, iz] * mol_dens,
                    )
                    for i in comps:
                        _set_if_unfixed(
                            self.mole_frac_comp[t, ix, iz, i],
                            self.conc_mol_comp[t, ix, iz, i] / mol_dens,
                        )
                    _set_if_unfixed(
                        self.enth_mol[t, ix, iz],
                        sum(
                            common._comp_enthalpy_expr(self.temperature[t, ix, iz], i)
                            * self.mole_frac_comp[t, ix, iz, i]
                            for i in comps
                        ),
                    )
                    if self.config.has_gas_holdup:
                        _set_if_unfixed(
                            self.int_energy_mol[t, ix, iz],
                            sum(
                                common._comp_int_energy_expr(
                                    self.temperature[t, ix, iz], i
                                )
                                * self.mole_frac_comp[t, ix, iz, i]
                                for i in comps
                            ),
                        )
                        _set_if_unfixed(
                            self.int_energy_density[t, ix, iz],
                            self.int_energy_mol[t, ix, iz] / self.vol_mol[t, ix, iz],
                        )
                for i in comps:
                    _set_if_unfixed(
                        self.conc_mol_comp_deviation_x1[t, iz, i],
                        self.conc_mol_comp_deviation_x[t, self.ixnodes.last(), iz, i],
                    )

        solver_obj = get_solver(solver, optarg)
        common._init_solve_block(self, solver_obj, solve_log)

    def calculate_scaling_factors(self):
        pass

    def model_check(self, steady_state=True):
        comp_set = set(self.component_list)
        elements_present = set()

        for element in _element_list:
            include_element = False
            for species in _gas_species_list:
                # Floating point equality take warning!
                if species in comp_set and _element_dict[element][species] != 0:
                    include_element = True
            if include_element:
                elements_present.add(element)

        if not steady_state:
            # Mass and energy conservation equations steady state only at present
            return
        for t in self.flowsheet().config.time:
            for element in _element_list:
                if element not in elements_present:
                    continue
                sum_in = 0
                sum_out = 0
                for iz in self.iznodes:
                    sum_in += sum(
                        _element_dict[element][j]
                        * self.material_flux_x0[t, iz, j]
                        * self.xface_area[iz]
                        for j in self.component_list
                    )
                    sum_out += sum(
                        _element_dict[element][j]
                        * self.material_flux_x1[t, iz, j]
                        * self.xface_area[iz]
                        for j in self.component_list
                    )
                normal = max(
                    pyo.value(sum_in), pyo.value(sum_out), 1e-8
                )  # FIXME justify this number
                fraction_change = pyo.value((sum_out - sum_in) / normal)
                if abs(fraction_change) > 3e-3:
                    raise RuntimeError(
                        f"{element} is not being conserved in {self.name}; "
                        f"fractional change {fraction_change}"
                    )
            enth_in = 0
            enth_out = 0

            for iz in self.iznodes:
                enth_in += self.xface_area[iz] * (
                    self.heat_flux_x0[t, iz]
                    + sum(
                        common._comp_enthalpy_expr(self.temperature_x0[t, iz], j)
                        * self.material_flux_x0[t, iz, j]
                        for j in self.component_list
                    )
                )
                enth_out += self.xface_area[iz] * (
                    self.heat_flux_x1[t, iz]
                    + sum(
                        common._comp_enthalpy_expr(self.temperature_x1[t, iz], j)
                        * self.material_flux_x1[t, iz, j]
                        for j in self.component_list
                    )
                )
            total_joule_heating = sum(
                sum(self.joule_heating[t, ix, iz] for ix in self.ixnodes)
                for iz in self.iznodes
            )

            normal = max(
                pyo.value(abs(enth_in)),
                pyo.value(abs(enth_out)),
                pyo.value(abs(total_joule_heating)),
                1e-4,
            )  # FIXME justify this number
            fraction_change = pyo.value(
                (enth_out - enth_in - total_joule_heating) / normal
            )
            if abs(fraction_change) > 3e-3:
                raise RuntimeError(
                    f"Energy is not being conserved in {self.name}; "
                    f"fractional change {fraction_change}"
                )

    def recursive_scaling(self):
        gsf = iscale.get_scaling_factor

        def ssf(c, s):
            iscale.set_scaling_factor(c, s, overwrite=False)

        sgsf = iscale.set_and_get_scaling_factor

        def cst(c, s):
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        sR = 1e-1  # Scaling factor for R
        sD = 1e5  # Heuristic scaling factor for diffusion coefficient
        sy_def = 10  # Mole frac comp scaling
        sH = 1e-4  # Enthalpy/int energy
        sk = 0.1  # Fudge factor to scale temperature_deviation_x
        sLx = sgsf(self.length_x, len(self.ixnodes) / self.length_x.value)
        sLy = 1 / self.length_y[None].value
        sLz = len(self.iznodes) / self.length_z[None].value

        for t in self.flowsheet().time:
            for iz in self.iznodes:
                if not self.temperature_z[t, iz].is_reference():
                    sT = sgsf(self.temperature_z[t, iz], 1e-2)

                if self.heat_flux_x0[t, iz].is_reference():
                    sq0 = gsf(self.heat_flux_x0[t, iz].referent, default=1e-2)
                else:
                    sq0 = sgsf(self.heat_flux_x0[t, iz], 1e-2)
                cst(self.heat_flux_x0_eqn[t, iz], sq0)
                if not self.temperature_deviation_x0.is_reference():
                    ssf(self.temperature_deviation_x0, sq0 * sLx / sk)

                if self.heat_flux_x1[t, iz].is_reference():
                    sq1 = gsf(self.heat_flux_x1[t, iz].referent, default=1e-2)
                else:
                    sq1 = sgsf(self.heat_flux_x1[t, iz], 1e-2)
                cst(self.heat_flux_x1_eqn[t, iz], sq1)
                if not self.temperature_deviation_x1.is_reference():
                    ssf(self.temperature_deviation_x1, sq1 * sLx / sk)

                sqx = min(sq0, sq1)
                # sqz = 10 * sqx  # Heuristic

                smaterial_flux_x = {}
                for j in self.component_list:
                    # ssf(self.conc_mol_comp_ref[t,iz,j],sy_def*1e-4/(sR*sT))

                    if self.material_flux_x0[t, iz, j].is_reference():
                        smaterial_flux_x0 = gsf(
                            self.material_flux_x0[t, iz, j].referent, default=1e-2
                        )
                    else:
                        smaterial_flux_x0 = sgsf(self.material_flux_x0[t, iz, j], 1e-2)
                    cst(self.material_flux_x0_eqn[t, iz, j], smaterial_flux_x0)
                    if not self.conc_mol_comp_deviation_x0[t, iz, j].is_reference():
                        ssf(
                            self.conc_mol_comp_deviation_x0[t, iz, j],
                            smaterial_flux_x0 * sLx / sD,
                        )

                    if self.material_flux_x1[t, iz, j].is_reference():
                        smaterial_flux_x1 = gsf(
                            self.material_flux_x1[t, iz, j].referent, default=1e-2
                        )
                    else:
                        smaterial_flux_x1 = sgsf(self.material_flux_x1[t, iz, j], 1e-2)
                    cst(self.material_flux_x1_eqn[t, iz, j], smaterial_flux_x1)

                    if not self.conc_mol_comp_deviation_x1[t, iz, j].is_reference():
                        ssf(
                            self.conc_mol_comp_deviation_x1[t, iz, j],
                            smaterial_flux_x1 * sLx / sD,
                        )

                    smaterial_flux_x[j] = min(smaterial_flux_x0, smaterial_flux_x1)

                for ix in self.ixnodes:
                    sP = sgsf(self.pressure[t, ix, iz], 1e-4)
                    sV = sR * sT / sP

                    sDT = sgsf(self.temperature_deviation_x[t, ix, iz], sqx * sLx / sk)
                    sH = sgsf(self.enth_mol[t, ix, iz], sH)
                    cst(self.enth_mol_eqn[t, ix, iz], sH)

                    if self.config.has_holdup:
                        if self.config.has_gas_holdup:
                            sU = sgsf(self.int_energy_mol[t, ix, iz], sH)
                            cst(self.int_energy_mol_eqn[t, ix, iz], sU)

                            s_rho_U = sgsf(self.int_energy_density[t, ix, iz], sU / sV)
                            cst(self.int_energy_density_eqn[t, ix, iz], s_rho_U)

                        s_rho_U_solid = sgsf(
                            self.int_energy_density_solid[t, ix, iz],
                            1
                            / (
                                self.solid_heat_capacity.value
                                * self.solid_density.value
                                * sDT
                            ),
                        )
                        cst(self.int_energy_density_solid_eqn[t, ix, iz], s_rho_U_solid)

                    cst(self.mole_frac_comp_eqn[t, ix, iz], 1)
                    cst(self.energy_balance_solid_eqn[t, ix, iz], sqx * sLy * sLz)

                    for j in self.component_list:
                        sy = sgsf(self.mole_frac_comp[t, ix, iz, j], sy_def)
                        cst(self.conc_mol_comp_eqn[t, ix, iz, j], sy * sP)
                        ssf(
                            self.conc_mol_comp_deviation_x[t, ix, iz, j],
                            smaterial_flux_x[j] * sLx / sD,
                        )

                        cst(
                            self.material_balance_eqn[t, ix, iz, j],
                            smaterial_flux_x[j] * sLy * sLz,
                        )
                        # FIXME come back later to clean up
                        ssf(self.diff_eff_coeff[t, ix, iz, j], 1e5)
                        cst(self.diff_eff_coeff_eqn[t, ix, iz, j], 1e5)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for SurrogateBlock
"""
import pytest
import re

from pyomo.environ import ConcreteModel, Set, Var

from idaes.core.surrogate.surrogate_block import SurrogateBlock, _extract_var_data


@pytest.mark.unit
def test_extract_var_data_None():
    assert _extract_var_data(None) is None


@pytest.mark.unit
def test_extract_var_data_scalar():
    m = ConcreteModel()
    m.v = Var()
    assert _extract_var_data(m.v) == [m.v]


@pytest.mark.unit
def test_extract_var_data_indexed():
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3, 4])
    m.v = Var(m.s)
    assert _extract_var_data(m.v) == [m.v[1], m.v[2], m.v[3], m.v[4]]


@pytest.mark.unit
def test_extract_var_data_indexed_not_ordered():
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3, 4], ordered=False)
    m.v = Var(m.s)
    with pytest.raises(
        ValueError, match="Expected IndexedVar: v to be indexed over " "an ordered set."
    ):
        _extract_var_data(m.v)


@pytest.mark.unit
def test_extract_var_data_multiple():
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3, 4])
    m.vs = Var()
    m.vi = Var(m.s)
    assert _extract_var_data([m.vs, m.vi, m.vs]) == [
        m.vs,
        m.vi[1],
        m.vi[2],
        m.vi[3],
        m.vi[4],
        m.vs,
    ]


@pytest.mark.unit
def test_extract_var_data_not_var():
    m = ConcreteModel()
    with pytest.raises(
        ValueError,
        match="Unknown variable type of <class "
        "'pyomo.core.base.PyomoModel.ConcreteModel'> for unknown",
    ):
        _extract_var_data(m)


@pytest.mark.unit
def test_setup_inputs_outputs_zero_inputs():
    m = ConcreteModel()
    m.sb = SurrogateBlock()
    with pytest.raises(
        ValueError,
        match="Attempting to create a Surrogate block with no inputs "
        "and/or no outputs. A SurrogateBlock must have at least "
        "one input and output",
    ):
        m.sb._setup_inputs_outputs(0, 1)


@pytest.mark.unit
def test_setup_inputs_outputs_zero_outputs():
    m = ConcreteModel()
    m.sb = SurrogateBlock()
    with pytest.raises(
        ValueError,
        match="Attempting to create a Surrogate block with no inputs "
        "and/or no outputs. A SurrogateBlock must have at least "
        "one input and output",
    ):
        m.sb._setup_inputs_outputs(1, 0)


@pytest.mark.unit
def test_setup_inputs_outputs_no_vars_or_labels():
    m = ConcreteModel()
    m.sb = SurrogateBlock()

    m.sb._setup_inputs_outputs(4, 5)

    assert m.sb._input_labels == [0, 1, 2, 3]
    assert isinstance(m.sb.inputs_set, Set)
    assert isinstance(m.sb.inputs, Var)
    for i in m.sb.inputs_set:
        assert i in m.sb._input_labels
        assert m.sb.inputs[i].value == 0
    assert m.sb._input_vars == [
        m.sb.inputs[0],
        m.sb.inputs[1],
        m.sb.inputs[2],
        m.sb.inputs[3],
    ]

    assert m.sb._output_labels == [0, 1, 2, 3, 4]
    assert isinstance(m.sb.outputs_set, Set)
    assert isinstance(m.sb.outputs, Var)
    for i in m.sb.outputs_set:
        assert i in m.sb._output_labels
        assert m.sb.outputs[i].value == 0
    assert m.sb._output_vars == [
        m.sb.outputs[0],
        m.sb.outputs[1],
        m.sb.outputs[2],
        m.sb.outputs[3],
        m.sb.outputs[4],
    ]


@pytest.mark.unit
def test_setup_inputs_outputs_input_label_mismatch():
    m = ConcreteModel()
    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match="Specifying input_labels to a SurrogateBlock, but the "
        "length does not match n_inputs",
    ):
        m.sb._setup_inputs_outputs(4, 5, input_labels=["x1"])


@pytest.mark.unit
def test_setup_inputs_outputs_output_label_mismatch():
    m = ConcreteModel()
    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match="Specifying output_labels to a SurrogateBlock, but the "
        "length does not match n_outputs",
    ):
        m.sb._setup_inputs_outputs(4, 5, output_labels=["x1"])


@pytest.mark.unit
def test_setup_inputs_outputs_w_labels_no_vars():
    m = ConcreteModel()
    m.sb = SurrogateBlock()

    m.sb._setup_inputs_outputs(
        4,
        5,
        input_labels=["x1", "x2", "x3", "x4"],
        output_labels=["z1", "z2", "z3", "z4", "z5"],
    )

    assert m.sb._input_labels == ["x1", "x2", "x3", "x4"]
    assert isinstance(m.sb.inputs_set, Set)
    assert isinstance(m.sb.inputs, Var)
    for i in m.sb.inputs_set:
        assert i in m.sb._input_labels
        assert m.sb.inputs[i].value == 0
    assert m.sb._input_vars == [
        m.sb.inputs["x1"],
        m.sb.inputs["x2"],
        m.sb.inputs["x3"],
        m.sb.inputs["x4"],
    ]

    assert m.sb._output_labels == ["z1", "z2", "z3", "z4", "z5"]
    assert isinstance(m.sb.outputs_set, Set)
    assert isinstance(m.sb.outputs, Var)
    for i in m.sb.outputs_set:
        assert i in m.sb._output_labels
        assert m.sb.outputs[i].value == 0
    assert m.sb._output_vars == [
        m.sb.outputs["z1"],
        m.sb.outputs["z2"],
        m.sb.outputs["z3"],
        m.sb.outputs["z4"],
        m.sb.outputs["z5"],
    ]


@pytest.fixture(scope="module")
def surr_block():
    m = ConcreteModel()
    m.input_set = Set(initialize=["x1", "x2", "x3", "x4"])
    m.output_set = ["z1", "z2", "z3", "z4", "z5"]
    m.inputs = Var(m.input_set)
    m.outputs = Var(m.output_set)

    m.sb = SurrogateBlock()

    m.sb._setup_inputs_outputs(4, 5, input_vars=m.inputs, output_vars=m.outputs)

    return m


@pytest.mark.unit
def test_setup_inputs_outputs_w_vars_no_labels(surr_block):
    assert surr_block.sb._input_labels == [0, 1, 2, 3]
    assert not hasattr(surr_block.sb, "inputs_set")
    assert not hasattr(surr_block.sb, "inputs")
    assert surr_block.sb._input_vars == [
        surr_block.inputs["x1"],
        surr_block.inputs["x2"],
        surr_block.inputs["x3"],
        surr_block.inputs["x4"],
    ]

    assert surr_block.sb._output_labels == [0, 1, 2, 3, 4]
    assert not hasattr(surr_block.sb, "outputs_set")
    assert not hasattr(surr_block.sb, "outputs")
    assert surr_block.sb._output_vars == [
        surr_block.outputs["z1"],
        surr_block.outputs["z2"],
        surr_block.outputs["z3"],
        surr_block.outputs["z4"],
        surr_block.outputs["z5"],
    ]


@pytest.mark.unit
def test_setup_inputs_outputs_input_vars_mismatch():
    m = ConcreteModel()
    m.input_set = Set(initialize=["x1", "x2", "x3"])
    m.output_set = ["z1", "z2", "z3", "z4", "z5"]
    m.inputs = Var(m.input_set)
    m.outputs = Var(m.output_set)

    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Specifying input_vars to a SurrogateBlock, but "
            "the length of the input_vars (after extracting all"
            " the VarData objects) does not match n_inputs"
        ),
    ):
        m.sb._setup_inputs_outputs(4, 5, input_vars=m.inputs, output_vars=m.outputs)


@pytest.mark.unit
def test_setup_inputs_outputs_input_vars_empty_list():
    m = ConcreteModel()
    m.output_set = ["z1", "z2", "z3", "z4", "z5"]
    m.outputs = Var(m.output_set)

    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Specifying input_vars to a SurrogateBlock, but "
            "the length of the input_vars (after extracting all"
            " the VarData objects) does not match n_inputs"
        ),
    ):
        m.sb._setup_inputs_outputs(4, 5, input_vars=[], output_vars=m.outputs)


@pytest.mark.unit
def test_setup_inputs_outputs_output_vars_mismatch():
    m = ConcreteModel()
    m.input_set = Set(initialize=["x1", "x2", "x3", "x4"])
    m.output_set = ["z1", "z2", "z3", "z4"]
    m.inputs = Var(m.input_set)
    m.outputs = Var(m.output_set)

    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Specifying output_vars to a SurrogateBlock, but "
            "the length of the output_vars (after extracting all"
            " the VarData objects) does not match n_outputs"
        ),
    ):
        m.sb._setup_inputs_outputs(4, 5, input_vars=m.inputs, output_vars=m.outputs)


@pytest.mark.unit
def test_setup_inputs_outputs_output_vars_empty_list():
    m = ConcreteModel()
    m.input_set = Set(initialize=["x1", "x2", "x3", "x4"])
    m.inputs = Var(m.input_set)

    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Specifying output_vars to a SurrogateBlock, but "
            "the length of the output_vars (after extracting all"
            " the VarData objects) does not match n_outputs"
        ),
    ):
        m.sb._setup_inputs_outputs(4, 5, input_vars=m.inputs, output_vars=[])


@pytest.mark.unit
def test_input_vars_as_list(surr_block):
    assert surr_block.sb._input_vars_as_list() == [
        surr_block.inputs["x1"],
        surr_block.inputs["x2"],
        surr_block.inputs["x3"],
        surr_block.inputs["x4"],
    ]


@pytest.mark.unit
def test_output_vars_as_list(surr_block):
    assert surr_block.sb._output_vars_as_list() == [
        surr_block.outputs["z1"],
        surr_block.outputs["z2"],
        surr_block.outputs["z3"],
        surr_block.outputs["z4"],
        surr_block.outputs["z5"],
    ]


@pytest.mark.unit
def test_input_vars_as_dict(surr_block):
    assert surr_block.sb.input_vars_as_dict() == {
        0: surr_block.inputs["x1"],
        1: surr_block.inputs["x2"],
        2: surr_block.inputs["x3"],
        3: surr_block.inputs["x4"],
    }


@pytest.mark.unit
def test_output_vars_as_dict(surr_block):
    assert surr_block.sb.output_vars_as_dict() == {
        0: surr_block.outputs["z1"],
        1: surr_block.outputs["z2"],
        2: surr_block.outputs["z3"],
        3: surr_block.outputs["z4"],
        4: surr_block.outputs["z5"],
    }


class DummySurrogate:
    # Dummy a SurrogateObject to use to test build method
    def populate_block(self, blk, **kwargs):
        blk._populate_called = True

    def n_inputs(self):
        return 4

    def n_outputs(self):
        return 5

    def input_labels(self):
        return ["x1", "x2", "x3", "x4"]

    def output_labels(self):
        return ["z1", "z2", "z3", "z4", "z5"]

    def input_bounds(self):
        return {"x1": (1, 10), "x2": (2, 20), "x3": (3, 30), "x4": (4, 40)}


@pytest.mark.unit
def test_build_model():
    m = ConcreteModel()

    m.dummy = DummySurrogate()

    m.sb = SurrogateBlock()

    m.sb.build_model(m.dummy, use_surrogate_bounds=False)

    # Check that the _populate_called attribute was added to the SurrogateBlock
    assert m.sb._populate_called

    # Check variables and bounds
    assert m.sb._input_labels == ["x1", "x2", "x3", "x4"]
    assert isinstance(m.sb.inputs_set, Set)
    assert isinstance(m.sb.inputs, Var)
    for i in m.sb.inputs_set:
        assert i in m.sb._input_labels
        assert m.sb.inputs[i].value == 0
        assert m.sb.inputs[i]._lb is None
        assert m.sb.inputs[i]._ub is None
    assert m.sb._input_vars == [
        m.sb.inputs["x1"],
        m.sb.inputs["x2"],
        m.sb.inputs["x3"],
        m.sb.inputs["x4"],
    ]

    assert m.sb._output_labels == ["z1", "z2", "z3", "z4", "z5"]
    assert isinstance(m.sb.outputs_set, Set)
    assert isinstance(m.sb.outputs, Var)
    for i in m.sb.outputs_set:
        assert i in m.sb._output_labels
        assert m.sb.outputs[i].value == 0
    assert m.sb._output_vars == [
        m.sb.outputs["z1"],
        m.sb.outputs["z2"],
        m.sb.outputs["z3"],
        m.sb.outputs["z4"],
        m.sb.outputs["z5"],
    ]


@pytest.mark.unit
def test_build_model_w_bounds():
    m = ConcreteModel()

    m.dummy = DummySurrogate()

    m.sb = SurrogateBlock()

    m.sb.build_model(m.dummy, use_surrogate_bounds=True)

    # Check that the _populate_called attribute was added to the SurrogateBlock
    assert m.sb._populate_called

    # Check variables and bounds
    assert m.sb._input_labels == ["x1", "x2", "x3", "x4"]
    assert isinstance(m.sb.inputs_set, Set)
    assert isinstance(m.sb.inputs, Var)
    for i in m.sb.inputs_set:
        assert i in m.sb._input_labels
        assert m.sb.inputs[i].value == 0
        assert m.sb.inputs[i]._lb == m.dummy.input_bounds()[i][0]
        assert m.sb.inputs[i]._ub == m.dummy.input_bounds()[i][1]
    assert m.sb._input_vars == [
        m.sb.inputs["x1"],
        m.sb.inputs["x2"],
        m.sb.inputs["x3"],
        m.sb.inputs["x4"],
    ]

    assert m.sb._output_labels == ["z1", "z2", "z3", "z4", "z5"]
    assert isinstance(m.sb.outputs_set, Set)
    assert isinstance(m.sb.outputs, Var)
    for i in m.sb.outputs_set:
        assert i in m.sb._output_labels
        assert m.sb.outputs[i].value == 0
    assert m.sb._output_vars == [
        m.sb.outputs["z1"],
        m.sb.outputs["z2"],
        m.sb.outputs["z3"],
        m.sb.outputs["z4"],
        m.sb.outputs["z5"],
    ]


@pytest.mark.unit
def test_build_model_unused_kwarg():
    m = ConcreteModel()

    m.dummy = DummySurrogate()

    m.sb = SurrogateBlock()

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Error in keyword arguments passed to "
            "build_model. The following arguments were not used: "
            "['foo']"
        ),
    ):
        m.sb.build_model(m.dummy, use_surrogate_bounds=False, foo=True)
